/*
 * Copyright (c) 2001
 * Stefan Saroiu and the University of Washington
 * 
 * Copyright (c) 1998, 1999
 * Stefan Savage and the University of Washington.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author(s) may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.  
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * HISTORY
 * 05-Apr-01  Stefan Saroiu (tzoompy) at the University of Washington
 *      sprobe-0.1 released. Can estimate bb from local to remote for any host
 *      and remote to local for gnutella only.
 *
 * 03-Apr-01  Stefan Saroiu (tzoompy) at the University of Washington
 *      Remove Sting code that dealt with packets losses.
 *      Add the SProbe bottleneck link bandwidth routines for downstream
 *        (local to remote) and upstream (remote to local) - Gnutella only
 *
 * 04-Nov-99  Stefan Savage (savage) at the University of Washington
 *	Port to Linux
 *
 * 03-Nov-99  Stefan Savage (savage) at the University of Washington
 *	Cleanup, removed overlapping big packets (bug)
 *
 * 01-Apr-99  Stefan Savage (savage) at the University of Washington
 *	Real request strings and limited receiver window.  AD tests.
 *
 * 21-Mar-99  Stefan Savage (savage) at the University of Washington
 *	Switched to OOO ack parity, overlapping big packets
 *
 * 21-Dec-98  Stefan Savage (savage) at the University of Washington
 *	Created.
 */

#include <stdio.h>
#include <math.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h> 
#ifdef __FreeBSD__
#include <netinet/ip_fw.h>
#endif /* __FreeBSD__ */
#ifdef linux
struct tcphdr {char x[20];};
struct iphdr {char x[20];};
struct udphdr {char x[20];};
struct icmphdr {char x[20];};
#include <linux/if.h>
#include <linux/ip_fw.h> 
#endif /* linux */
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include "inet.h"
#include "capture.h"
#include "setsignal.h"

#define DEFAULT_COUNT		(1)
#define DEFAULT_TARGETPORT	(80)
#define DEFAULT_PACKETSIZE	(1460)
#define DEFAULT_TRAINLENGTH     (6)

char *defaultRequest;
char defaultRequestFormat[]="GET %s HTTP/1.0\nAccept: text/plain\nAccept: */*\nUser-Agent: Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt; Sting)\n\n";

#define SYNTIMEOUT		(3.0)	/* 3 seconds */
#define MAXSYNRETRANSMITS	(3)
#define MAXDATARETRANSMITS	(5)

#define ACKSENDERTIMEOUT        (3)     /* 1 second */
#define MAXACKSENDERRETRANSMITS (3)
#define MAXPSHRETRANSMITS       (3)
#define PSHTIMEOUT              (3)
#define GNUTELLAWAITTIMEOUT     (10)    /* how long to wait in gnutella */

char *progName;

extern char *optarg;	/* Variables used for getopt(3) */
extern int optind;
extern int opterr;
extern int optopt;

struct TcpSession session;
int acksReceived = 0;
int dataSent = 0;
int dataSeenSent = 0;
int dataReceived = 0;
int acksSent = 0;
int dataLost = 0;
int acksLost = 0;

int trainlength = DEFAULT_TRAINLENGTH;
int downstream = 0;
int verbose = 0;
int interactive = 0;
int initSession = 0;
int initCapture = 0;
int initFirewall = 0;
int killOldFlows = 1;
int timeOut = SYNTIMEOUT;

#ifdef __FreeBSD__
struct ip_fw firewallRule;
#endif /* __FreeBSD__ */
#ifdef linux
struct ip_fwchange firewallRule;
#endif /* linux */

/*
 * Send three resets to the receiver to cleanup its state.
 */
void SendReset(struct TcpSession *s)
{
  struct TcpPacket p;
  int i;

  for (i=0;i<3;i++) {  /* 3 is totally arbitrary */
    SendSessionPacket(s, &p, 0, 0, TCPFLAGS_RST);
  }
}

void RespondWithReset(struct TcpPacket *p) 
{
  struct TcpSession s;

  memset(&s, 0, sizeof(struct TcpSession));
  printf("Resetting connection\n");
  s.src = p->ip.ip_dst;
  s.dst = p->ip.ip_src;
  s.dport = ntohs(p->tcp.tcp_sport);
  s.sport = ntohs(p->tcp.tcp_dport);
  s.snd_nxt = ntohl(p->tcp.tcp_ack);
  s.rcv_nxt = ntohl(p->tcp.tcp_seq);
  SendReset(&s);
}


/* make a clean exit on interrupts */
RETSIGTYPE
Cleanup(int signo)
{


  /* If a firewall rule has been installed then remove it */
  if (initFirewall > 0) {
#ifdef linux
#define IP_FW_DEL	(IP_FW_DELETE)
#endif /* linux */
    if (setsockopt(session.socket,IPPROTO_IP, IP_FW_DEL,
		   &firewallRule, sizeof(firewallRule)) != 0) {
      printf("ERROR: couldn't remove firewall rule\n");
    }
  }

  if (initSession > 0) {
    SendReset(&session);
    shutdown(session.socket,2);
  }
  
  if (initCapture > 0) {
    CaptureEnd();
  }

  exit(0);
}

void Quit()
{
  fflush(stdout);
  fflush(stderr);
  Cleanup(1);
  exit(-1);
}

void Usage() 
{
  printf("usage:\n%s:",progName);
  printf("\t [-p target-port] [-c count] [-s data-size]\n");
  printf("\t [-w source-port] [-v] [-t train-length]\n");
  printf("\t [-d] [-h] [-g]\n");
  Quit();
}

int BindTcpPort(int sockfd)
{
   struct sockaddr_in  sockName;
   int port, result;
   int randomOffset;

#define START_PORT (10*1024) 
#define END_PORT   (0xFFFF)

   /* Choose random offset to reduce likelihood of collision with last run */
   randomOffset = (int)(1000.0*drand48());

   /* Try to find a free port in the range START_PORT+1..END_PORT */
   port = START_PORT+randomOffset;
   do {
       ++port;
       sockName.sin_addr.s_addr = INADDR_ANY;       
       sockName.sin_family = AF_INET;
       sockName.sin_port = htons(port);
       result = bind(sockfd, (struct sockaddr *)&sockName,
                     sizeof(sockName));
   } while ((result < 0) && (port < END_PORT));
   
   if (result < 0) {
     /* No free ports */
       perror("bind");
       port = 0;
   }   
   return port;
}


void PrintTimeStamp(struct timeval *ts)
{
	(void)printf("%02d:%02d:%02d.%06u ",
		     (unsigned int)ts->tv_sec / 3600,
		     ((unsigned int)ts->tv_sec % 3600) / 60,
		     (unsigned int)ts->tv_sec % 60, (unsigned int)ts->tv_usec);
}

double GetTime()
{
  struct timeval tv;
  struct timezone tz;
  double postEpochSecs;

  if (gettimeofday(&tv, &tz) < 0) {
    perror("GetTime");
    exit(-1);
  }
  
  postEpochSecs = (double)tv.tv_sec + ((double)tv.tv_usec/(double)1000000.0);
  return postEpochSecs;
}

/*
 * Does a packet belong to a directional flow in our session?
 */
#define INSESSION(p, src, sport, dst, dport)			\
    (((p)->ip.ip_src == (src)) && ((p)->ip.ip_dst == (dst)) &&	\
     ((p)->ip.ip_p == IPPROTOCOL_TCP) &&			\
     ((p)->tcp.tcp_sport == htons(sport)) &&			\
     ((p)->tcp.tcp_dport == htons(dport)))

/*
 * Initialize firewall, socket, etc.. This code is common to all SProbe
 * routines
 */
void
Init(uint32 sourceAddress, uint16 sourcePort,
     uint32 targetAddress, uint16 targetPort,
     struct TcpSession *s)
{
  int rawSocket;
  int flag=1;

  /* Initialize session structure */
  memset(s, 0, sizeof(struct TcpSession));
  s->src = sourceAddress;
  s->sport = sourcePort;
  s->dst = targetAddress;
  s->dport = targetPort; 
  s->snd_wnd = 0;
  s->rcv_nxt = 0;
  s->irs = 0;

  /* Now open a raw socket for sending our "fake" TCP segments */
  if ((rawSocket = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    printf("ERROR: couldn't open socket\n");
    Quit();
  }
  if (setsockopt(rawSocket, IPPROTO_IP, IP_HDRINCL,
		 (char *)&flag,sizeof(flag)) < 0) {
    printf("ERROR: couldn't set raw socket options\n");
    Quit();
  }

  s->socket = rawSocket;
  
  /* 
   * Block kernel TCP from seeing any of the target's responses.
   * AFAIK its irrelvant that this is the same raw socket that we're
   * sending on.
   *
   * I wish that everyone would get their act together and support
   * ONE firewall API... imagine that.
   */

  memset(&firewallRule, 0, sizeof(firewallRule));

#ifdef __FreeBSD__
  firewallRule.fw_flg |= IP_FW_F_DENY | IP_FW_F_IN; 
  firewallRule.fw_prot = IPPROTO_TCP;
  firewallRule.fw_src.s_addr = s->dst;
  firewallRule.fw_smsk.s_addr = htonl(~0);
  firewallRule.fw_uar.fw_pts[0] = s->dport;
  IP_FW_SETNSRCP(&firewallRule, 1);
  firewallRule.fw_dst.s_addr = s->src;
  firewallRule.fw_uar.fw_pts[1] = s->sport;
  IP_FW_SETNDSTP(&firewallRule, 1);
  firewallRule.fw_dmsk.s_addr = htonl(~0);
  firewallRule.fw_number = 999;
  if (setsockopt(s->socket, IPPROTO_IP, IP_FW_ADD,
		 &firewallRule, sizeof(firewallRule)) != 0) {
    printf("ERROR: couldn't block kernel TCP for %s:%d\n",
	   InetAddress(s->dst), s->dport);
  }
#endif /* __FreeBSD__ */
#ifdef linux
  memcpy(firewallRule.fwc_label, IP_FW_LABEL_INPUT,
	 sizeof(firewallRule.fwc_label));
  memcpy(firewallRule.fwc_rule.label, IP_FW_LABEL_BLOCK,
	 sizeof(firewallRule.fwc_label)); 
  firewallRule.fwc_rule.ipfw.fw_proto = IPPROTO_TCP;

  firewallRule.fwc_rule.ipfw.fw_src.s_addr = s->dst;
  firewallRule.fwc_rule.ipfw.fw_smsk.s_addr = htonl(~0); 
  firewallRule.fwc_rule.ipfw.fw_spts[0] = s->dport;
  firewallRule.fwc_rule.ipfw.fw_spts[1] = s->dport;
  firewallRule.fwc_rule.ipfw.fw_dst.s_addr = s->src;
  firewallRule.fwc_rule.ipfw.fw_dmsk.s_addr = htonl(~0); 
  firewallRule.fwc_rule.ipfw.fw_dpts[0] = s->sport;
  firewallRule.fwc_rule.ipfw.fw_dpts[1] = s->sport;

  if (setsockopt(s->socket,IPPROTO_IP,IP_FW_APPEND,
		 &firewallRule,sizeof(firewallRule)) != 0) {
    printf("ERROR: couldn't block kernel TCP for %s:%d\n",
	   InetAddress(s->dst),  s->dport);
    Quit();
  }
#endif /* linux */
  initFirewall=1;
}

/*
 * This is some common code that gets called a lot. This function checks
 * whether the packet has the Reset or FIN bit on, and if so, prints a
 * message and terminate a connection.
 * When a packet a received the caller should check whether the packet
 * contains useful information (it's a PSH for example) and after it's done
 * call this function to check RST and FIN in the packet
 */
void
CheckResetOrFin(struct TcpPacket *p, char *msg) {
  /* Is it a RST?  i.e., please stop */
  if (p->tcp.tcp_flags & TCPFLAGS_RST) { 
    printf("WARNING: Reset encountered during %s\n", msg);
    Quit();
  }
  /* We can also get FINs, especially if we confused the other stack
   * in a previous run of SProbe? This is rare but it can happen */
  else if (p->tcp.tcp_flags & TCPFLAGS_FIN) { 
    printf("WARNING: FINs from the other host, while %s\n", msg);
    Quit();
  }
  /* No? then its something weird */
  else {
    printf("WARNING: unexpected packet during %s\n", msg);
    PrintTcpPacket(p);
  }
}

/*
 * This function does the Tcp handshake - SYN, SYN/ACK, ACK.
 * It takes a pointer to the TcpSession, some memory allocated for the packet
 * and the lengths of the options and data fragments in the packet being sent.
 * Here, datalen should be 0 mostly.
 */
void
TcpHandshake(struct TcpSession *s, struct TcpPacket *synPacket, int optlen,
	     int datalen)
{
  struct TcpPacket *p;
  struct PacketInfo pi;
  double timeoutTime;
  int synAckReceived = 0;
  int numRetransmits = 0;

  /* Send the SYN - first step in the TCP connection */
  SendSessionPacket(s, synPacket, optlen, datalen, TCPFLAGS_SYN);  

  /* 
   * Wait for SYN/ACK and retransmit SYN if appropriate 
   * Kinda crappy, but it gets the job done 
   */
  timeoutTime = GetTime() + SYNTIMEOUT;
  while(!synAckReceived && numRetransmits < MAXSYNRETRANSMITS) {
    while(GetTime() < timeoutTime) {

      /* Have we captured any packets? */
      if ((p = (struct TcpPacket *)CaptureGetPacket(&pi)) == NULL) {
	continue;
      }
      
      /* Received a packet from us to them */
      if (INSESSION(p, s->src, s->sport, s->dst, s->dport)) {
	/* Is it a SYN? If so this is our syn packet */
	if (p->tcp.tcp_flags == TCPFLAGS_SYN) {
	  continue;
	}
	/* No? then its something weird */
	else {
	  printf("WARNING: unexpected packet during session establishment\n");
	  PrintTcpPacket(p);
	  continue;
	}
      }
      /* Next, check if a packet from them to us */
      else if (INSESSION(p, s->dst, s->dport, s->src, s->sport)) {      
	/* Is it a SYN/ACK? */
	if (p->tcp.tcp_flags == (TCPFLAGS_SYN | TCPFLAGS_ACK)) {
	  synAckReceived = 1;
	  break;
	}
	/* Is it a RST?  i.e., please stop */
	else CheckResetOrFin(p, "tcp handshake");
      }
      else {
	printf("WARNING: unexpected packet during session establishment\n");
	PrintTcpPacket(p);
	if (killOldFlows) {
	  /* If it was sent to us and wasn't a RST then tell them to stop */
	  if (p->ip.ip_dst == s->src && !(p->tcp.tcp_flags & TCPFLAGS_RST)) {
	    RespondWithReset(p);
	  }
	}
      }
    } /* while(GetTime() < timeoutTime) */
    if (!synAckReceived) {
      if (verbose) {
	printf("SYN timeout.  Retransmitting\n");
      }
      SendSessionPacket(s, synPacket, optlen, datalen, TCPFLAGS_SYN);
      timeoutTime = GetTime() + SYNTIMEOUT;
      numRetransmits++;
    }
  } /* while */
  if (numRetransmits >= MAXSYNRETRANSMITS) {
    printf("Could not establish contact after %d retries\n", numRetransmits);
    Quit();
  }

  /* Extract the MSS */
  if(p->tcp.tcp_hl > 80) {
    int ip_hl = ((0x0f & p->ip.ip_vhl) << 2);
    int offset, limit = ip_hl + (p->tcp.tcp_hl >> 2);
    for(offset = ip_hl + 20; offset < limit; offset += 1) { 
      if(((unsigned char*)p)[offset] == 0 ||
	 ((unsigned char*)p)[offset] == 1) continue;
      if(((unsigned char*)p)[offset] == 2) {
	limit = (ntohs(*((int*)(((unsigned char*)p) + offset + 2))));
	if(limit == 1460) 
	  printf("MSS: 1460 suggesting bandwidth > 128Kbps\n");
	else if(limit == 536)
	  printf("MSS: 536 suggesting bandwidth < 128Kbps (modem perhaps?)\n");
	else
	  printf("MSS: %d\n", limit);
	break;
      }
      offset += 2;
    }
  }
    
    
  
  /* Update session variables */
  s->irs = ntohl(p->tcp.tcp_seq); /* receiver sequence number */
  s->rcv_nxt = s->irs+1;  /* SYN/ACK takes up a byte of seq space */
  s->snd_nxt = s->iss+1;  /* SYN takes up a byte of seq space */
  s->snd_una = s->iss+1;  

  /* send the sender ack to the TCP handshake */
  s->iss = s->snd_nxt;
  SendSessionPacket(s, synPacket, optlen, datalen, TCPFLAGS_ACK);  
  timeoutTime = GetTime() + ACKSENDERTIMEOUT;
  /* 
   * Wait for SYN/ACK and retransmit SYN if appropriate 
   * Kinda crappy, but it gets the job done 
   */
  while(GetTime() < timeoutTime) {
    /* Have we captured any packets? */
    if ((p = (struct TcpPacket *)CaptureGetPacket(&pi)) == NULL) {
      continue;
    }

    /* Received a packet from us to them */
    if (INSESSION(p, s->src, s->sport, s->dst, s->dport)) {
      /* Is it an ACK? If so it's ours. */
      if (p->tcp.tcp_flags == TCPFLAGS_ACK) {
	initSession = 1;
	break;
      }
      /* No? then its something weird */
      else {
	printf("WARNING: unexpected packet during session establishment\n");
	PrintTcpPacket(p);
	continue;
      }
    }
    /* Next, check if a packet from them to us */
    else if (INSESSION(p, s->dst, s->dport, s->src, s->sport)) {
      CheckResetOrFin(p, "waiting to see our ack of TcpHandshake");
    }
  }
  
  /* If we haven't seen our own packet, something it's weird. quit */
  if(!initSession) {
    printf("We tried to ack the handshake, but our packet hasn't been sent\n");
    Quit();
  }
}

/*
 * This function abstract the Gnutella handshake. In this handshake, we
 * send a "GNUTELLA CONNECT" message and we're waiting to see the "GNUTELLA OK"
 * Notice that this application handshake will occur even if the party
 * does not accept anymore gnutella connections. It will send us a FIN later
 * revealing his unwillingness to have anymore connections open
 */
void
GnutellaHandshake(struct TcpSession *s, struct TcpPacket *synPacket,
		  unsigned char *buffPacket, int optlen, int datalen,
		  int optoffset)
{
  struct TcpPacket *p;
  struct PacketInfo pi;
  double timeoutTime;
  int numRetransmits = 0;
  int pshReceived = 0;
  int ackReceived = 0;

  /* This is ugly, but it gets the job done. Put the "GNUTELLA CONNECT" string
   * in the packet to be sent */
  memcpy(buffPacket + optoffset, "GNUTELLA CONNECT/0.4\n\n", 22);
  s->snd_una = s->snd_nxt + 22;

  while(!ackReceived &&	numRetransmits < MAXACKSENDERRETRANSMITS) {
    /* Let's send it baby, do the same timeout tricks as before */
    SendSessionPacket(s, synPacket, optlen, datalen,
		      TCPFLAGS_ACK | TCPFLAGS_PSH);

    /* We need to wait now to see their ack */
    timeoutTime = GetTime() + ACKSENDERTIMEOUT;  
    while(GetTime() < timeoutTime) {
      /* Have we captured any packets? */
      if ((p = (struct TcpPacket *)CaptureGetPacket(&pi)) == NULL) {
	continue;
      }

      /* Received a packet from us to them */
      if (INSESSION(p, s->src, s->sport, s->dst, s->dport)) {
	/* Is this our packet? */
	if (p->tcp.tcp_flags == (TCPFLAGS_ACK | TCPFLAGS_PSH)) {
	  continue;		    /* seems like our packet */
	}
	/* No? then its something weird */
	else {
	  printf("WARNING: unexpected packet during session establishment\n");
	  PrintTcpPacket(p);
	  continue;
	}
      }
      /* Next, check if a packet from them to us */
      else if (INSESSION(p, s->dst, s->dport, s->src, s->sport)) {      
	/* Is this an ack to our packet ? */
	if (p->tcp.tcp_flags & TCPFLAGS_ACK) {
	  if(ntohl(p->tcp.tcp_ack) == s->snd_una) {
	    ackReceived = 1;
	    if(p->tcp.tcp_flags & TCPFLAGS_PSH) pshReceived = 1;
	    break;
	  }
	}
	else CheckResetOrFin(p, "gnutella handshake");
      }
    } /* while(GetTime() < timeoutTime) */
    numRetransmits += 1;
  } /* while */
  if(!ackReceived) {
    printf("Timeout after %d retries for getting an ack to gnutella handshake\n", MAXACKSENDERRETRANSMITS);
    Quit();
  }

  /* We just only got the ack, we need to wait for GNUTELLA OK now */
  if(!pshReceived) {
    timeoutTime = GetTime() + ACKSENDERTIMEOUT;
    while(GetTime() < timeoutTime) {
      /* Have we captured any packets? */
      if ((p = (struct TcpPacket *)CaptureGetPacket(&pi)) == NULL) {
	continue;
      }

      /* Next, check if a packet from them to us */
      else if (INSESSION(p, s->dst, s->dport, s->src, s->sport)) {      
	/* Is this a psh ? */
	if (p->tcp.tcp_flags == (TCPFLAGS_ACK | TCPFLAGS_PSH)) {
	  if(!strncmp(&(((unsigned char*)p)[20 + (p->tcp.tcp_hl >> 2)]),
		      "GNUTELLA OK\n\n", 13)){
	    pshReceived = 1;
	    break;
	  }
	  else {
	    printf("A PSH message but without the GNUTELLA OK messsage in it\n");
	    Quit();
	  }
	}
	else CheckResetOrFin(p, "receival of GNUTELLA OK");
      }
    } /* while */
    if(!pshReceived) {
      printf("No PSH message received while waiting for GNUTELLA OK\n");
      Quit();    
    }
  } /* if */

  /* While we're here, we received the PSH, let's reflect that */
  s->irs = ntohl(p->tcp.tcp_seq);
  s->rcv_nxt += ntohs(p->ip.ip_len) - ((0x0f & p->ip.ip_vhl) << 2) -
    (p->tcp.tcp_hl >> 2);
  s->snd_una = ntohl(p->tcp.tcp_ack); /* the unack byte is the same as the
				         party's ack */
  s->snd_nxt = s->snd_una;

  /* Notice that we haven't acked the psh yet */
}

void
Latencies(uint32 sourceAddress, uint16 sourcePort,
	  uint32 targetAddress, uint16 targetPort,
	  struct TcpSession *s)
{
  struct TcpPacket *p;
  struct PacketInfo pi;
  double timeoutTime;
  struct TcpPacket synPacket;
  double trst = -1;			    /* stores the times of resets */
  double tsyn = -1;			    /* stores the times of syns */

  Init(sourceAddress, sourcePort, targetAddress, targetPort, s);
  s->snd_nxt = (uint32)mrand48();  /* random initial sequence number */
  s->iss = s->snd_nxt;
  SendSessionPacket(s, &synPacket, 0, 0, TCPFLAGS_SYN);
  s->snd_nxt += 1;
  timeoutTime = GetTime() + trainlength * SYNTIMEOUT;
  while(GetTime() < timeoutTime && (tsyn == -1 || trst == -1)) {
    /* Have we captured any packets? */
    if ((p = (struct TcpPacket *)CaptureGetPacket(&pi)) == NULL) {
      continue;
    }
    /* Received a packet from us to them */
    if (INSESSION(p, s->src, s->sport, s->dst, s->dport)) {
      /* Is it a SYN? */
      if (p->tcp.tcp_flags == TCPFLAGS_SYN) {
	tsyn = pi.ts.tv_sec + (double)pi.ts.tv_usec/1000000.0;
	continue;
      }
      else {
	printf("WARNING: unexpected packet sent to the host\n");
	PrintTcpPacket(p);
	continue;
      }
    }
    else /* packets from them to us */ {
      if (INSESSION(p, s->dst, s->dport, s->src, s->sport)) {
	trst = pi.ts.tv_sec + (double)pi.ts.tv_usec/1000000.0;
	continue;
      }
    }
  }
  if(-1 != trst && -1 != tsyn) 
    printf("Latency: %.6f ms\n", 1000 * (trst - tsyn));
  else
    printf("Latency unknown\n");
}

/*
 * Here we use our 'brains' to get the downstream estimate. The trick is
 * simple. Send two fat SYN packets and measure the time dispersion of the
 * RSTs received. A Fat SYN packet is a SYN packet (40 bytes TCP/IP header) +
 * 1460 bytes of payload, which is has 0 effect other than making the packet
 * fat.
 */
void
DownstreamEstimate(uint32 sourceAddress, uint16 sourcePort,
		   uint32 targetAddress, uint16 targetPort,
		   struct TcpSession *s, int count, int dataSize)
{
  struct TcpPacket *p;
  struct PacketInfo pi;
  double timeoutTime, delta;
  int i, k;
  unsigned char *buff = NULL;
  struct TcpPacket *synPacket = NULL;
  double *trst;			    /* stores the times of resets */
  double *tsyn;			    /* stores the times of syns */
  int crst;			    /* count resets */
  int csyn;			    /* count syns */
  FILE *hcd;			    /* handler to the hcd.dat */
  int iRST = (trainlength - 1) / 2; /* index of reset of fat syn 1 */
  int iiRST = iRST + 1;             /* index of reset of fat syn 2 */
  int rstss = 0;		    /* rst sequence number */
  double fatltn;		    /* fat latency */
  double sknltn;		    /* skinny latency */
  /*
   * The data we add as payload must be highly uncompressable. Why ? - to
   * get rid of any modem compression schemes. Such data can be found in
   * hcd.dat. We have precisely 1500 bytes there, read that in the buffer
   */
  hcd = fopen("./hcd.dat", "rb");
  if(NULL == hcd) {
    printf("The file hcd.dat couldn't be open.\n");
    printf("SProbe cannot estimate downstream bandwidths.\n");
    exit(-1);
  }
  if(verbose) printf("hcd.dat opened\n");

  /* Ugly hack, allocate memory and then case the first bytes for the packet */
  if(NULL ==
     (buff = (unsigned char*) malloc(sizeof(struct TcpPacket) + dataSize))) {
    printf("ERROR: Run out of memory\n");
    exit(-1);
  }
  synPacket = (struct TcpPacket *)buff;
  
  if(1500 != fread(buff, 1, 1500, hcd)) {
    printf("SProbe couldn't read 1500 bytes of data from hcd.dat\n");
    exit(-1);
  }
  if(verbose) printf("1500 bytes read\n");
  fclose(hcd);

  Init(sourceAddress, sourcePort, targetAddress, targetPort, s);

  if(NULL == (trst = (double*)malloc(trainlength * sizeof(double))) ||
     NULL == (tsyn = (double*)malloc(trainlength * sizeof(double)))) {
    printf("Error: Out of memory\n");
    exit(-1);
  }

  /*
   * SProbe tries count times to get a good measurement. If not, too bad,
   * but at least we say I don't know
   */
  for(k = 0; k < count; k += 1) {
    s->snd_nxt = (uint32)mrand48();  /* random initial sequence number */
    s->iss = s->snd_nxt;
    rstss = 0;
    
    /*
     * Send the packet pair and then send the remaining of the train
     */
    for(i = 2; i < trainlength; i += 2) {
      SendSessionPacket(s, synPacket, 0, 0, TCPFLAGS_SYN);
      s->snd_nxt += 1;
    }
    SendSessionPacket(s, synPacket, 0, dataSize, TCPFLAGS_SYN);
    s->snd_nxt += dataSize + 1;
    SendSessionPacket(s, synPacket, 0, dataSize, TCPFLAGS_SYN);
    s->snd_nxt += dataSize + 1;
    for(i = 3; i < trainlength; i += 2) {
      SendSessionPacket(s, synPacket, 0, 0, TCPFLAGS_SYN);
      s->snd_nxt += 1;
    }

    bzero(trst, sizeof(double) * trainlength);
    bzero(tsyn, sizeof(double) * trainlength);
    crst = 0; csyn = 0;
    timeoutTime = GetTime() + trainlength * SYNTIMEOUT;
    while(GetTime() < timeoutTime) {
      /* Have we captured any packets? */
      if ((p = (struct TcpPacket *)CaptureGetPacket(&pi)) == NULL) {
	continue;
      }
      /* Received a packet from us to them */
      if (INSESSION(p, s->src, s->sport, s->dst, s->dport)) {
	/* Is it a SYN? */
	if (p->tcp.tcp_flags == TCPFLAGS_SYN) {
	  /* Great one of our syns */
	  tsyn[csyn++] = pi.ts.tv_sec + (double)pi.ts.tv_usec/1000000.0;
	  continue;
	}
	else {
	  printf("WARNING: unexpected packet sent to the host\n");
	  PrintTcpPacket(p);
	  continue;
	}
      }
      else /* packets from them to us */
	if (INSESSION(p, s->dst, s->dport, s->src, s->sport)) {
	  if(!(p->tcp.tcp_flags & TCPFLAGS_RST)) {
	    printf("Warning: The host responded with packets other than RSTs\n");
	    printf("Try a different port\n");
	    Quit();
	  }
	  trst[crst++] = pi.ts.tv_sec + (double)pi.ts.tv_usec/1000000.0;
	  if(ntohl(p->tcp.tcp_ack) <= rstss) {
	    printf("RSTs came back out of order\n");
	    CaptureSkipNPackets(trainlength - crst, &pi);
	    crst = -1;
	    break;
	  }
	  else if(0 != rstss && ntohl(p->tcp.tcp_ack) - rstss > dataSize + 1) {
	    printf("RSTs acknowledges %d bytes\n",(int)
		   (ntohl(p->tcp.tcp_ack) - (u_long)rstss));
	    CaptureSkipNPackets(trainlength - crst, &pi);
	    crst = -1;
	    break;
	  }
	  rstss = ntohl(p->tcp.tcp_ack);
	  if(crst == trainlength) break; /* check for termination */
	}
    }
    if(crst == trainlength){
      delta = trst[iiRST] - trst[iRST];

      /* Compute the latencies */
      if(0 != tsyn[iiRST] && 0 != tsyn[iRST])
	fatltn = (trst[iiRST] - tsyn[iiRST] + trst[iRST] - tsyn[iRST]) / 2;
      else
	fatltn = -1;
      
      sknltn = 0;
      for(i = 0; i < trainlength; i += 1) {
	if(i == iRST || i == iiRST) continue;
	if(0 == tsyn[i])
	  break;
	sknltn += trst[i] - tsyn[i]; 
      }
      if(i == trainlength)
	sknltn /= (trainlength - 2);
      else
	sknltn = -1;
      
      /* Check whether RSTs of FAT SYNs are closer than any other RST pair */
      for(i = 1; i < trainlength; i += 1) {	
	if(i == iRST || i == iiRST || i == iiRST + 1) continue;
	
	if(trst[i] - trst[i-1] > delta) {
	  printf("RST of FAT SYNs (%d %d) are closer than another RST (%d %d) pair (bb: %.3fKbps, rtt(fat):%.6f, rtt(skn): %6f)\n", iRST, iiRST, i-1, i, 8 * (dataSize + sizeof(struct TcpPacket)) / delta / 1000.0, fatltn, sknltn);
	  break;
	}
      }

      /* IF the RSTs are spaced well, report, and terminate */
      if(i == trainlength) {
	printf("Estimate: %.3fKbps (rtt(fat): %.6f, rtt(skn): %.6f)\n",
	       8 * (dataSize + sizeof(struct TcpPacket)) / delta / 1000.0,
	       fatltn, sknltn);
	return;
      }
    }
    else if(crst > 0) {
      printf("SProbe sent %d SYNs and received %d RSTs\n", trainlength, crst);
      printf("Either: \n");
      printf("\t-host did not reply with a RST for each SYN, or...\n");
      printf("\t-SYNs were lost\n");
    }
    else if(crst == 0) {
      printf("Host is either offline or does not reset any packets. Bummer.\n");
      return;
    }
    if(verbose) printf("Sending three resets...\n");
#if 0				    /* no need to */
    SendReset(s);
    CaptureSkipNPackets(3, &pi);
#endif
    if(k + 1 < count) sleep(5);	    /* do not bombard the other guy */
  }
  
  printf("SProbe tried %d times to get an estimate but cross-traffic got in the way.\n", count);
  printf("Rather than fooling you, SProbe gives up. Have a nice day !\n");
  Quit();
}

/*
 * This function, construct a properly formatted Get request to be sent to the
 * server to which we want to measure bb to.
 */
void
ConstructWWWGetRequest(char *target)
{
  char *loc;			    /* points to the first '/' if any */
  
  /* Yucky, but fast way of getting rid of http:// */
  if((target[0] == 'h' || target[0] == 'H') &&
     (target[1] == 't' || target[1] == 'T') &&
     (target[2] == 't' || target[2] == 'T') &&
     (target[3] == 'p' || target[3] == 'P') &&
     (target[4] == ':') &&
     (target[5] == '/') &&
     (target[6] == '/'))

    target = &(target[7]);

  /* Get some space for the get request */
  defaultRequest = (char*)malloc(strlen(defaultRequestFormat) + 256);
  if(NULL == defaultRequest) {
    printf("Error: Out of memory\n");
    exit(-1);
  }

  /* We need to separate the directory and filename of the request from the
   * name of the server. ie separate www.abc.com/dir/file.jpg into www.abc.com
   * and /dir/file.jpg. The directory and filename go into the get request
   */
  loc = strchr(target, '/');
  if(NULL != loc) {
    sprintf(defaultRequest, defaultRequestFormat, loc);
    *loc = '\0';
  }
  else				    /* otherwise, just add a / to get req. */
    sprintf(defaultRequest, defaultRequestFormat, "/");  
}

/*
 * Estimate upstream bandwidth for Gnutella application 
 */
void
GnutellaEstimate(uint32 sourceAddress, uint16 sourcePort,
		 uint32 targetAddress, uint16 targetPort,
		 struct TcpSession *s)
{
  unsigned char *buffPacket;
  struct TcpPacket *synPacket;
  short mss = 1460;
  int optoffset;
  int numPackets = 0;
  int ack0Received = 0;
  struct TcpPacket *p;
  struct PacketInfo pi;
  double timeoutTime;
  int numRetransmits;
  int length1, length2;		    /* packet lengths of the packet pair */
  double ts1, ts2;
  
  /* Initialize packet to TCP/IP header + MSS option */
  if((buffPacket =
      (unsigned char*)malloc(sizeof(struct TcpPacket) + 8 + 40)) ==
     NULL) {
    printf("ERROR: out of memory\n");
    Quit();
  }
  
  /* Construct a MSS of 1460 */
  optoffset = sizeof(struct TcpPacket);
  buffPacket[optoffset++] = 2;
  buffPacket[optoffset++] = 4;
  mss = htons(mss);
  memcpy(&(buffPacket[optoffset]), &mss, 2);
  optoffset += 2;
  buffPacket[optoffset++] = 1;
  buffPacket[optoffset++] = 3;
  buffPacket[optoffset++] = 3;
  buffPacket[optoffset++] = 0;

  (unsigned char*)synPacket = buffPacket;
  
  Init(sourceAddress, sourcePort, targetAddress, targetPort, s);
  s->snd_wnd = 20;		    /* advertise a small window, big enough
				       to do the gnutella handshakes */
  s->snd_nxt = (uint32)mrand48();  /* random initial sequence number */
  s->iss = s->snd_nxt;

  TcpHandshake(s, synPacket, 8, 0); /* do the handshake */
  /* Do the gnutella handshake. Note that this function, waits for the
   * GNUTELLA OK PSH packet from the party, but doesn't ack it.
   */
  GnutellaHandshake(s, synPacket, buffPacket, 8, 22, optoffset);
  
  /* Now we need to do the magic. We're going to advertise a rwin of 0 to
   * the party and wait some time. After that we'll send two duplicate acks
   * which in return should trigger him into sending two large packets.
   * If this succeeeds we're home free.
   */  
  s->snd_wnd = 0;

  /* We're going to ack the received push. Since we're wasting a packet on
   * this anyway, we'll send a gnutella ping packet too, so that we show
   * our eagerness to received packets
   */
  /* Build the gnutella ping */
  /* The first 16 bytes are some id, we'll just leave the bytes in place */
  buffPacket[optoffset + 16] = 0;   /* next is the message type, 0 for ping */
  buffPacket[optoffset + 17] = 10;  /* TTL */
  bzero(&(buffPacket[optoffset + 18]), 5); /* the rest are 0s */

  numRetransmits = 0;
  while(!ack0Received && numRetransmits < MAXACKSENDERRETRANSMITS) {
    SendSessionPacket(s, synPacket, 8, 23, TCPFLAGS_ACK | TCPFLAGS_PSH);
    timeoutTime = GetTime() + PSHTIMEOUT;

    while(GetTime() < timeoutTime) {
      /* Have any packets  ? */
      if ((p = (struct TcpPacket *)CaptureGetPacket(&pi)) == NULL) {
	continue;
      }

      /* Received a packet from us to them */
      if (INSESSION(p, s->src, s->sport, s->dst, s->dport)) {
	/* Is it our packet ? */
	if (p->tcp.tcp_flags == (TCPFLAGS_ACK | TCPFLAGS_PSH)) {
	  continue;		    /* our packet probably */
	}
	/* No? then its something weird */
	else {
	  printf("WARNING: unexpected packet while waiting for ack to rwin 0\n");
	  PrintTcpPacket(p);
	  continue;
	}
      }
      /* Next, check if a packet from them to us */
      else if (INSESSION(p, s->dst, s->dport, s->src, s->sport)) {
	if(p->tcp.tcp_flags == TCPFLAGS_ACK) {
	  ack0Received = 1;
	  break;
	}
	else CheckResetOrFin(p, "while sending our ack with rwin 0\n");
      }    
    } /* while(GetTime() < timeoutTime) */
    numRetransmits += 1;
  } /* while */
  if(!ack0Received) {
    printf("The party did not ack our Gnutella Ping (with rwin=0)\n");
    Quit();
  }
  
  /* While we're here, let's update the state */
  s->irs = ntohl(p->tcp.tcp_seq);
  s->rcv_nxt += ntohs(p->ip.ip_len) - ((0x0f & p->ip.ip_vhl) << 2) -
    (p->tcp.tcp_hl >> 2);
  s->snd_una = ntohl(p->tcp.tcp_ack); /* the unack byte is the same as the
				         party's ack */
  s->snd_nxt = s->snd_una;

  /* We need to wait some time now. The other party might start sending
   * window probing packets now. Let's consume them with pcap
   */
  timeoutTime = GetTime() + GNUTELLAWAITTIMEOUT;
  while(GetTime() < timeoutTime) {
    /* Have we captured any packets? */
    if ((p = (struct TcpPacket *)CaptureGetPacket(&pi)) == NULL) {
      continue;
    }

    /* If a packet from them to us */
    if (INSESSION(p, s->dst, s->dport, s->src, s->sport)) {      
      if(p->tcp.tcp_flags == TCPFLAGS_ACK) {
	continue;		    /* probing packet */
      }
      else CheckResetOrFin(p, "during waiting for buffers to fill up\n");
    }
    else {
      printf("WARNING: unexpected packet while waiting for buffers to fill up\n");
      PrintTcpPacket(p);
      continue;
    }
  } /* while(GetTime() < timeoutTime) */

  s->snd_wnd = 32120;  
  s->snd_wnd = 32120;
  /* We're sending there ACK's - (Solaris needs 3) */
  SendSessionPacket(s, synPacket, 8, 0, TCPFLAGS_ACK);
  SendSessionPacket(s, synPacket, 8, 0, TCPFLAGS_ACK);

  timeoutTime = GetTime() + 2 * ACKSENDERTIMEOUT; /* wait two timeouts */
  while(numPackets < 2 && GetTime() < timeoutTime) {
    /* Have we captured any packets? */
    if ((p = (struct TcpPacket *)CaptureGetPacket(&pi)) == NULL) {
      continue;
    }
      /* Next, check if a packet from them to us */
    if (INSESSION(p, s->dst, s->dport, s->src, s->sport)) {      
      /* Is it a SYN/ACK? */
      if (p->tcp.tcp_flags & TCPFLAGS_FIN) {
	printf("Not accepting connection\n");
	return;
      }
      else if(p->tcp.tcp_flags == (TCPFLAGS_ACK | TCPFLAGS_PSH)) {
	if(0 == numPackets) {	    /* if first packet */
	  length1 = ntohs(p->ip.ip_len);
	  ts1 = pi.ts.tv_sec + (double)pi.ts.tv_usec/1000000.0;
	}
	else {
	  length2 = ntohs(p->ip.ip_len);
	  ts2 = pi.ts.tv_sec + (double)pi.ts.tv_usec/1000000.0;
	}
	printf("Packet received w/ length: %d at %.6f\n", ntohs(p->ip.ip_len),
	       pi.ts.tv_sec + (double)pi.ts.tv_usec/1000000.0);
	numPackets += 1;
      }
      else CheckResetOrFin(p, "waiting for the packet pair\n");
    }
  }
  if(numPackets < 2) {
    printf("Did not received two large packets from the party\n");
    return;
  }
  if(length1 < 1000 || length2 < 1000 || (length1 < 1300 && length2 < 1300)) {
    printf("The packets received aren't large enough. Try again\n");
    return;
  }
  printf("Estimate: %.3fKbps\n", 1500 * 8 / (ts2 - ts1) / length1);
  return;
}
   

/*
 * Setup a TCP session between our machine and the specified target
 * Extract and return the key bits of TCP state
 */
int EstablishSession(uint32 sourceAddress, uint16 sourcePort,
		     uint32 targetAddress, uint16 targetPort,
		     struct TcpSession *s)
{
  int rawSocket;
  struct TcpPacket *p;
  unsigned char *buffPacket;
  struct TcpPacket *synPacket;
  struct TcpPacket ackPacket;
  struct PacketInfo pi;
  int synAckReceived = 0;
  int ackSent = 0;
  int numRetransmits = 0;
  double timeoutTime;
  double ts1, ts2;
  int flag=1;
  short mss = 1460;
  int optoffset;

  /* Initialize packet to TCP/IP header + MSS option */
  if((buffPacket = (unsigned char*)malloc(sizeof(struct TcpPacket) + 8)) ==
     NULL) {
    printf("ERROR: out of memory\n");
    Quit();
  }
  
  /* Construct a MSS of 1460 */
  optoffset = sizeof(struct TcpPacket);
  buffPacket[optoffset++] = 2;
  buffPacket[optoffset++] = 4;
  mss = htons(mss);
  memcpy(&(buffPacket[optoffset]), &mss, 2);
  optoffset += 2;
  buffPacket[optoffset++] = 1;
  buffPacket[optoffset++] = 3;
  buffPacket[optoffset++] = 3;
  buffPacket[optoffset] = 0;

  (unsigned char*)synPacket = buffPacket;
  
  /* Initialize session structure */
  memset(s, 0, sizeof(struct TcpSession));
  s->src = sourceAddress;
  s->sport = sourcePort;
  s->dst = targetAddress;
  s->dport = targetPort;
  s->snd_wnd = 32120;
  s->snd_nxt = (uint32)mrand48();  /* random initial sequence number */
  s->iss = s->snd_nxt;
  s->rcv_nxt = 0;
  s->irs = 0;

  /* Now open a raw socket for sending our "fake" TCP segments */
  if ((rawSocket = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    printf("ERROR: couldn't open socket\n");
    Quit();
  }
  if (setsockopt(rawSocket, IPPROTO_IP, IP_HDRINCL,
		 (char *)&flag,sizeof(flag)) < 0) {
    printf("ERROR: couldn't set raw socket options\n");
    Quit();
  }

  s->socket = rawSocket;

  /* 
   * Block kernel TCP from seeing any of the target's responses.
   * AFAIK its irrelvant that this is the same raw socket that we're
   * sending on.
   *
   * I wish that everyone would get their act together and support
   * ONE firewall API... imagine that.
   */

  memset(&firewallRule, 0, sizeof(firewallRule));

#ifdef __FreeBSD__
  firewallRule.fw_flg |= IP_FW_F_DENY | IP_FW_F_IN; 
  firewallRule.fw_prot = IPPROTO_TCP;
  firewallRule.fw_src.s_addr = s->dst;
  firewallRule.fw_smsk.s_addr = htonl(~0);
  firewallRule.fw_uar.fw_pts[0] = s->dport;
  IP_FW_SETNSRCP(&firewallRule, 1);
  firewallRule.fw_dst.s_addr = s->src;
  firewallRule.fw_uar.fw_pts[1] = s->sport;
  IP_FW_SETNDSTP(&firewallRule, 1);
  firewallRule.fw_dmsk.s_addr = htonl(~0);
  firewallRule.fw_number = 999;
  if (setsockopt(s->socket, IPPROTO_IP, IP_FW_ADD,
		 &firewallRule, sizeof(firewallRule)) != 0) {
    printf("ERROR: couldn't block kernel TCP for %s:%d\n",
	   InetAddress(s->dst), s->dport);
  }
#endif /* __FreeBSD__ */
#ifdef linux
  memcpy(firewallRule.fwc_label, IP_FW_LABEL_INPUT,
	 sizeof(firewallRule.fwc_label));
  memcpy(firewallRule.fwc_rule.label, IP_FW_LABEL_BLOCK,
	 sizeof(firewallRule.fwc_label)); 
  firewallRule.fwc_rule.ipfw.fw_proto = IPPROTO_TCP;

  firewallRule.fwc_rule.ipfw.fw_src.s_addr = s->dst;
  firewallRule.fwc_rule.ipfw.fw_smsk.s_addr = htonl(~0); 
  firewallRule.fwc_rule.ipfw.fw_spts[0] = s->dport;
  firewallRule.fwc_rule.ipfw.fw_spts[1] = s->dport;
  firewallRule.fwc_rule.ipfw.fw_dst.s_addr = s->src;
  firewallRule.fwc_rule.ipfw.fw_dmsk.s_addr = htonl(~0); 
  firewallRule.fwc_rule.ipfw.fw_dpts[0] = s->sport;
  firewallRule.fwc_rule.ipfw.fw_dpts[1] = s->sport;

  if (setsockopt(s->socket,IPPROTO_IP,IP_FW_APPEND,
		 &firewallRule,sizeof(firewallRule)) != 0) {
    printf("ERROR: couldn't block kernel TCP for %s:%d\n",
	   InetAddress(s->dst),  s->dport);
    Quit();
  }
#endif /* linux */
  initFirewall=1;

  /* Send SYN + an MSS of 1460 */
  SendSessionPacket(s, synPacket, 8, 0, TCPFLAGS_SYN);
  timeoutTime = GetTime() + SYNTIMEOUT;

  /* 
   * Wait for SYN/ACK and retransmit SYN if appropriate 
   * Kinda crappy, but it gets the job done 
   */
  while(!synAckReceived && numRetransmits < MAXSYNRETRANSMITS) {
    while(GetTime() < timeoutTime) {

      /* Have we captured any packets? */
      if ((p = (struct TcpPacket *)CaptureGetPacket(&pi)) == NULL) {
	continue;
      }

      /* Received a packet from us to them */
      if (INSESSION(p, s->src, s->sport, s->dst, s->dport)) {
	/* Is it a SYN? */
	if (p->tcp.tcp_flags == TCPFLAGS_SYN) {
	  ts1 = pi.ts.tv_sec + (double)pi.ts.tv_usec/1000000.0;
	  continue;
	}
	/* No? then its something weird */
	else {
	  printf("WARNING: unexpected packet during session establishment\n");
	  PrintTcpPacket(p);
	  continue;
	}
      }
      /* Next, check if a packet from them to us */
      else if (INSESSION(p, s->dst, s->dport, s->src, s->sport)) {      
	/* Is it a SYN/ACK? */
	if (p->tcp.tcp_flags == (TCPFLAGS_SYN | TCPFLAGS_ACK)) {
	  int z;
	  timeoutTime = GetTime(); /* force exit */
	  synAckReceived++;
	  ts2 = pi.ts.tv_sec + (double)pi.ts.tv_usec/1000000.0;
	  for(z = 0; z + 20 < (p->tcp.tcp_hl >> 2); z += 1) {
	    if(2 == ((unsigned char*)p)[40 + z]) {
	      printf("MSS is: %d\n", ((unsigned char*)p)[43+z] +
		     ((unsigned char*)p)[z+42] * 256);
	      break;
	    }
	  } 
	  if (numRetransmits > 0) {      
	    printf("(Unreliable) ");  /* ACK for which SYN? */
	  }
	  printf("Connection setup took %d ms\n",(int)((ts2-ts1)*1000.0));

	  /* Save ttl for,admittedly poor,indications of reverse route change */
	  s->ttl = p->ip.ip_ttl;
	  s->rcv_wnd = ntohl(p->tcp.tcp_win);
	}
	/* Is it a RST?  i.e., please stop */
	else if (p->tcp.tcp_flags & TCPFLAGS_RST) { 
	  printf("WARNING: Early Reset during session establishment\n");
	  Quit();
	}
	/* No? then its something weird */
	else {
	  printf("WARNING: unexpected packet during session establishment\n");
	  PrintTcpPacket(p);
	  continue;
	}
      }
      else {
	printf("WARNING: unexpected packet during session establishment\n");
	PrintTcpPacket(p);
	if (killOldFlows) {
	  /* If it was sent to us and wasn't a RST then tell them to stop */
	  if (p->ip.ip_dst == s->src && !(p->tcp.tcp_flags & TCPFLAGS_RST)) {
	    RespondWithReset(p);
	  }
	}
	continue;
      }
    }
    if (!synAckReceived) {
      if (verbose) {
	printf("SYN timeout.  Retransmitting\n");
      }
      SendSessionPacket(s, synPacket, 8, 0, TCPFLAGS_SYN);
      timeoutTime = GetTime() + SYNTIMEOUT;
      numRetransmits++;
    }
  }
  if (numRetransmits >= MAXSYNRETRANSMITS) {
    printf("Could not establish contact after %d retries\n", numRetransmits);
    Quit();
  } 
  
  /* Update session variables */
  s->irs = ntohl(p->tcp.tcp_seq); /* receiver sequence number */
  s->rcv_nxt = s->irs+1;  /* SYN/ACK takes up a byte of seq space */
  s->snd_nxt = s->iss+1;  /* SYN takes up a byte of seq space */
  s->snd_una = s->iss+1;  

  initSession = 1;

  /*
   * send the sender ack to the TCP handshake
   */
  s->iss = s->snd_nxt;
  SendSessionPacket(s, &ackPacket, 0, 0, TCPFLAGS_ACK);
  timeoutTime = GetTime() + ACKSENDERTIMEOUT;
  numRetransmits = 0;
  
  /* 
   * Wait for SYN/ACK and retransmit SYN if appropriate 
   * Kinda crappy, but it gets the job done 
   */
  while(!ackSent && numRetransmits < MAXACKSENDERRETRANSMITS) {
    while(GetTime() < timeoutTime) {
      /* Have we captured any packets? */
      if ((p = (struct TcpPacket *)CaptureGetPacket(&pi)) == NULL) {
	continue;
      }

      /* Received a packet from us to them */
      if (INSESSION(p, s->src, s->sport, s->dst, s->dport)) {
	/* Is it an ACK? */
	if (p->tcp.tcp_flags == TCPFLAGS_ACK) {
	  ackSent = 1;
	  break;
	}
	/* No? then its something weird */
	else {
	  printf("WARNING: unexpected packet during session establishment\n");
	  PrintTcpPacket(p);
	  continue;
	}
      }
      if(!ackSent && verbose) {
	printf("Sending ACK timeout. Retransmitting\n");
      }
      SendSessionPacket(s, &ackPacket, 0, 0, TCPFLAGS_ACK);
      timeoutTime = GetTime() + ACKSENDERTIMEOUT;
      numRetransmits += 1;
    }
    if(numRetransmits >= MAXACKSENDERRETRANSMITS) {
      printf("Couldn't send an ACK after %d retries\n", numRetransmits);
      Quit();
    }
  }

  printf("src = %s:%d (%u)\n", InetAddress(s->src),
	 s->sport, s->iss);
  printf("dst = %s:%d (%u)\n",InetAddress(s->dst),
	 s->dport, s->irs);

  free(buffPacket);
  return 1;
}

void
SendHTTPRequest(struct TcpSession *s)
{
  unsigned char *buff;
  struct TcpPacket *p;
  struct TcpPacket *pshPacket;
  double timeoutTime;
  struct PacketInfo pi;
  int numRetransmits = 0;
  int pshAckReceived = 0;
  int pshReceived = 0;
  double ts1, ts2;
  double ps, pps, ps1, ps2;
  double cwnd = 3;
  
  /* Initialize the get request (defaultRequest should not be more than 256) */
  if((buff = (unsigned char*)malloc(sizeof(struct TcpPacket) + 256)) == NULL) {
    printf("ERROR: Out of memory\n");
    Quit();    
  }
  /* append the payload to the end of the packet */
  memcpy(&(buff[sizeof(struct TcpPacket)]), defaultRequest,
	 strlen(defaultRequest));
  (unsigned char*)pshPacket = buff;

  /* Send out get request. Append an ack to increase receiver window to 2 */
  SendSessionPacket(s, pshPacket, 0, strlen(&(buff[sizeof(struct TcpPacket)])),
		    TCPFLAGS_ACK | TCPFLAGS_PSH);

  timeoutTime = GetTime() + PSHTIMEOUT;
  while(!pshAckReceived && numRetransmits < MAXPSHRETRANSMITS) {
    while(GetTime() < timeoutTime) {
      /* Have we captured any packets? */
      if ((p = (struct TcpPacket *)CaptureGetPacket(&pi)) == NULL) {
	continue;
      }

      /* Received a packet from us to them */
      if (INSESSION(p, s->src, s->sport, s->dst, s->dport)) {
	/* Is it a PSH? */
	if (p->tcp.tcp_flags == (TCPFLAGS_PSH | TCPFLAGS_ACK)) {
	  /* Great !!! I got my get request */
	  continue;
	}
	/* No? then its something weird */
	else {
	  printf("WARNING: unexpected packet during sending Get Request\n");
	  PrintTcpPacket(p);
	  continue;
	}
      }
      /* Next, check if a packet from them to us */
      else if (INSESSION(p, s->dst, s->dport, s->src, s->sport)) {
	/* Is it a ACK */
	if (p->tcp.tcp_flags == TCPFLAGS_ACK) {	  
	  /* Check whether the ack is for our packet - lots of garbage out
	   * there
	   */
	  if(s->iss + strlen(&buff[sizeof(struct TcpPacket)]) !=
	     ntohl(p->tcp.tcp_ack)) continue;

	  if(ntohs(p->ip.ip_len) >
	     ((0x0f & p->ip.ip_vhl) << 2) - (p->tcp.tcp_hl >> 2)) {
	    /*
	     * perhaps this is a push too (broken TCP)
	     */
	    printf("Assuming 'invisible' push\n");
	    pshReceived += 1;
	  }
	  pshAckReceived++;
	  if (numRetransmits > 0) {      
	    printf("(Unreliable) "); 
	  }
	  break;		    /* exit the inner while-loop */
	}
	else if (p->tcp.tcp_flags == (TCPFLAGS_ACK | TCPFLAGS_PSH)) {	  
	  /* Check whether the ack is for our packet - lots of garbage out
	   * there
	   */
	  if(s->iss + strlen(&buff[sizeof(struct TcpPacket)]) !=
	     ntohl(p->tcp.tcp_ack)) continue;

	  pshAckReceived++;
	  pshReceived ++;
	  ps = p->tcp.tcp_seq;
	  if (numRetransmits > 0) {      
	    printf("(Unreliable) "); 
	  }
	  break;		    /* exit the inner while-loop */
	}
	/* Is it a RST?  i.e., please stop */
	else if (p->tcp.tcp_flags & TCPFLAGS_RST) { 
	  printf("WARNING: Reset during Get Request\n");
	  Quit();
	}
	/* No? then its something weird */
	else {
	  printf("WARNING: unexpected packet during Get Request\n");
	  PrintTcpPacket(p);
	  continue;
	}
      }
      else {
	printf("WARNING: unexpected packet during Get Request\n");
	PrintTcpPacket(p);
	if (killOldFlows) {
	  /* If it was sent to us and wasn't a RST then tell them to stop */
	  if (p->ip.ip_dst == s->src && !(p->tcp.tcp_flags & TCPFLAGS_RST)) {
	    RespondWithReset(p);
	  }
	}
	continue;
      }
    }
    if(!pshAckReceived) {
      if (verbose) {
	printf("PSH timeout (Get Request).  Retransmitting\n");
      }
      SendSessionPacket(s, pshPacket, 0,
			strlen(&(buff[sizeof(struct TcpPacket)])),
			TCPFLAGS_ACK | TCPFLAGS_PSH);
      timeoutTime = GetTime() + PSHTIMEOUT;
      numRetransmits++;
    }
  }
  if (numRetransmits >= MAXPSHRETRANSMITS) {
    printf("Could not send get request after %d retries\n", numRetransmits);
    Quit();
  } 

  /* Update session variables */
  s->irs = ntohl(p->tcp.tcp_seq);
  s->snd_nxt = s->iss + strlen(&buff[sizeof(struct TcpPacket)]);

  /*
   * There are two cases. The server acked our push (with the get request).
   * First, it can append its push with the ack or the ack can be clear.
   * If the ack is alone, we need to wait for its push
   */
  timeoutTime = GetTime() + PSHTIMEOUT;
  /* if no push received w/ ack,wait for it*/
  while(0 == pshReceived && GetTime() < timeoutTime) {
    /* Have we captured any packets? */
    if ((p = (struct TcpPacket *)CaptureGetPacket(&pi)) == NULL) {
      continue;
    }
    /* Received a packet from us to them */
    if (INSESSION(p, s->src, s->sport, s->dst, s->dport)) {
      printf("WARNING: unexpected packet while waiting for their push\n");
      PrintTcpPacket(p);
      continue;
    }
    /* Next, check if a packet from them to us */
    else if (INSESSION(p, s->dst, s->dport, s->src, s->sport)) {
      /* Is it a PSH */
      if (p->tcp.tcp_flags == (TCPFLAGS_PSH | TCPFLAGS_ACK)) {
	ps = p->tcp.tcp_seq;
	pshReceived += 1;
      }	
    }
  }

  if(0 == pshReceived) {
    printf("Error: No Push received from the other end\n");
    Quit();
  }
  
  s->rcv_nxt = ntohl(p->tcp.tcp_seq) + ntohs(p->ip.ip_len) -
    ((0x0f & p->ip.ip_vhl) << 2) - (p->tcp.tcp_hl >> 2);

  // tzoompy - don't need it, do I
  //  SendSessionPacket(s, pshPacket, 0, 0, TCPFLAGS_ACK);
  
  /*
   * They acked our push request. Now they will send two large pushes
   * (hopefully). The trick is to measure those and then wait for them to
   * retransmit the first push. Only then we send an ack.
   */
  while(1) {
    pshReceived = 0;
    ts1 = -1;
    timeoutTime = GetTime() + cwnd * 2 * PSHTIMEOUT;
    while(pshReceived < cwnd && GetTime() < timeoutTime) {
      /* Have we captured any packets? */
      if ((p = (struct TcpPacket *)CaptureGetPacket(&pi)) == NULL) {
	continue;
      }
    
      /* Received a packet from us to them */
      if (INSESSION(p, s->src, s->sport, s->dst, s->dport)) {
	if(p->tcp.tcp_flags == TCPFLAGS_ACK) {
	  /* This is our ack, fine */
	  continue;
	}
	else {
	  printf("WARNING: unexpected packet during sending Get Request\n");
	  PrintTcpPacket(p);
	  continue;
	}
      }
      /* Next, check if a packet from them to us */
      else if (INSESSION(p, s->dst, s->dport, s->src, s->sport)) {
	/* Is it a PSH, sometimes the push flag is unset, the packet seems
	 * to be an ack only, but it carries data.
	 */
	if ((p->tcp.tcp_flags & TCPFLAGS_PSH) ||
	    (p->tcp.tcp_flags == TCPFLAGS_ACK)) {

	  pshReceived += 1;
	  pps = ntohl(p->tcp.tcp_seq) + ntohs(p->ip.ip_len) -
	    ((0x0f & p->ip.ip_vhl) << 2) - (p->tcp.tcp_hl >> 2);
	  if(pps > s->rcv_nxt) s->rcv_nxt = pps;
	  else {
	    /* a duplicate push packet received. This usually means, that
	     * we should resend an ack and try a new loop fresh
	     */
	    cwnd -= 1; break;
	  }
	  
	  ts2 = pi.ts.tv_sec + (double)pi.ts.tv_usec/1000000.0;
	  ps2 = ntohs(p->ip.ip_len);
	} 
	/* No? then its something weird */
	else if(!(p->tcp.tcp_flags & TCPFLAGS_FIN)) {
	  printf("WARNING: unexpected packet during data transfer\n");
	  PrintTcpPacket(p);
	}
	if(p->tcp.tcp_flags & TCPFLAGS_FIN) {
	  printf("WARNING: No more data. %d bytes received\n",
		 s->rcv_nxt -s->irs);
	  Quit();	
	}
	else if (p->tcp.tcp_flags & TCPFLAGS_RST) { 
	  printf("WARNING: Reset during receiving data\n");
	  Quit();
	}
      }
      if(-1 != ts1 && ps1 == 1500 && ps2 == 1500) {
/* 	This prints a ton more useful information, but, for release we */
/*	comment it out */
/*  	printf("ts1:%f, ts2:%f, ps1:%.1f, ps2:%.1f, bb:%.3fKbps\n", */
/*  	       ts1, ts2, ps1, ps2, 8.0 * ps1/(ts2 - ts1)/1000.0); */
	printf("Estimate: %.3fKbps!!!\n", 8.0 * ps1/(ts2 - ts1)/1000.0);
	Quit();
      }
      else if(-1 != ts1) 
//	printf("XXts1:%f, ts2:%f, ps1:%.1f, ps2:%.1f, bb:%.3fKbps\n",
//	       ts1, ts2, ps1, ps2, 8.0 * ps1/(ts2 - ts1)/1000.0);

      s->snd_nxt = ntohl(p->tcp.tcp_ack);
      ts1 = ts2;
      ps1 = ps2;
    }
    SendSessionPacket(s, pshPacket, 0, 0, TCPFLAGS_ACK);
    cwnd += 1;
  }
  Quit();
}

int
GetCannonicalInfo(char *string, char name[MAXHOSTNAMELEN], uint32 *address)
{
  struct hostent *hp;

  /* Is string in dotted decimal format? */
  if ((*address = inet_addr(string)) == INADDR_NONE) {
    /* No, then lookup IP address */
    if ((hp = gethostbyname(string)) == NULL) {
      /* Can't find IP address */
      printf("ERROR: couldn't obtain address for %s\n", string);
      return -1;
    } else {
      strncpy(name, hp->h_name, MAXHOSTNAMELEN-1);
      memcpy((void *)address, (void *)hp->h_addr, hp->h_length);
    }
  } else {
    if ((hp = gethostbyaddr((char *)address, sizeof(*address),
			    AF_INET)) == NULL) {
      /* Can't get cannonical hostname, so just use input string */
      if (verbose) {
	printf("WARNING: couldn't obtain cannonical name for %s\n", string);
      }
      strncpy(name, string, MAXHOSTNAMELEN - 1);
    } else {
      strncpy(name, hp->h_name, MAXHOSTNAMELEN - 1);
    }
  }
  return 0;
}

int main(int argc, char **argv) 
{
  int count = DEFAULT_COUNT;
  uint16 targetPort = DEFAULT_TARGETPORT;
  uint16 sourcePort = 0;
  int packetSize = DEFAULT_PACKETSIZE;

  char targetHostName[MAXHOSTNAMELEN];	/* DNS name of target host */
  uint32 targetIpAddress;		/* IP address of target host */
  char sourceHostName[MAXHOSTNAMELEN];
  uint32 sourceIpAddress;
  char *target;
  char source[MAXHOSTNAMELEN];
  char *p;
  struct sockaddr_in saddr;
  int fd;
  int opt;
  RETSIGTYPE (*oldhandler)(int);
  
  
  if ((p = (char *)strrchr(argv[0], '/')) != NULL) {
    progName = p + 1;
  } else {
    progName = argv[0];
  }

  opterr = 0;
  while ((opt = getopt(argc, argv, "p:c:s:w:t:vdghlf:")) != EOF) {
    switch (opt) {
    case 'p':
      targetPort = atoi(optarg);
      break;
    case 'w':
      sourcePort = atoi(optarg);
      break;
    case 'c':
      count = atoi(optarg);
      break;
    case 'l':
      downstream = -3;
      break;
    case 's':     
      packetSize = atoi(optarg);
      if (packetSize < 0 || packetSize > 1460) {
	printf("Valid data sizes are in the range 0-1460\n");
	Usage();
      }
      break; 
    case 'v':
      verbose = 1;
      break;
    case 'd':		    /* downstream, relative to us */
      downstream = 1;
      break;
    case 'h':
      downstream = -1;	    /* upstream, relative to us, Web Server */
      break;
    case 'g':
      downstream = -2;	    /* upstream, relative to us, Gnutella */
      break;
    case 'f':
#ifdef linux
      if(NULL == (stdout = fopen(optarg, "w"))) {
#endif
#ifdef __FreeBSD__
      if(NULL == freopen(optarg, "w", stdout)) {
#endif
	
	printf("Couldn't open output file: %s\n", argv[optind]);
	Usage();
      }
      break;
    case 't':
      trainlength = atoi(optarg);
      if(trainlength < 2 || trainlength > 100) {
	printf("Valid train lengths are in the range 2-100\n");
	Usage();
      }
      break;
    default:
      Usage();
    }
  }

  if(0 ==  interactive) {
    switch (argc - optind) {
    case 1:
      target = argv[optind];
      break;
    default:
      Usage();
    }
  }

  /* Check whether a direction was specified, either up or downstream */
  if(0 == downstream) {
    printf("Must specify a direction for your measurement.\n");
    printf("Use -g or -h for upstream and -d for downstream\n");
    Usage();
  }

  if(-1 == downstream)		    /* if http request needed upstream */
    /* Construct the get request to be sent to the web server */
    ConstructWWWGetRequest(target);

  /* Setup signal handler for cleaning up */
  (void)setsignal(SIGTERM, Cleanup);
  (void)setsignal(SIGINT, Cleanup);
  if ((oldhandler = setsignal(SIGHUP, Cleanup)) != SIG_DFL) 
    (void)setsignal(SIGHUP, oldhandler);

  printf("===================================================\n");
    
  /*
   * Get hostname and IP address of target host
   */
  if (GetCannonicalInfo(target, targetHostName, &targetIpAddress) < 0) {
    Quit();
  }
    
  printf("%s - %s - %.3f\n", target, targetHostName, GetTime());
  /*
   * Get hostname and IP address of source host
   */
  if (gethostname(source, MAXHOSTNAMELEN) != 0) {
    printf("ERROR: can't determine local hostname\n");
    Quit();
  }

  if (GetCannonicalInfo(source, sourceHostName, &sourceIpAddress) < 0) {
    Quit();
  }

  if (verbose) {
    printf("source = %s [%s]\n", sourceHostName,
	   InetAddress(sourceIpAddress));
    printf("target = %s [%s],\n",targetHostName,
	   InetAddress(targetIpAddress));
    printf("target port = %d, count = %d\n", targetPort, count);	   
  }

  /* Initialze random number generator */
  srand48(GetTime());
  if (sourcePort == 0) {
    /* 
     * Find and allocate a spare TCP port to use
     */
    saddr.sin_family = AF_INET;
    if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      printf("ERROR: can't open socket\n");
      return 0;
    }
    if ((sourcePort = BindTcpPort(fd)) == 0) {
      printf("ERROR: can't bind port\n");
      return 0;
    }
  }

  if(verbose) {
    printf("source port = %d\n", sourcePort);
  }
  
  /* Init packet capture device and install filter for our flow */
  CaptureInit(sourceIpAddress, sourcePort, targetIpAddress, targetPort);
  initCapture = 1;

  /*
   * if downstream, use the fat syn tricks
   */
  if(1 == downstream) {
    DownstreamEstimate(sourceIpAddress, sourcePort, targetIpAddress,
		       targetPort, &session, count, packetSize);
    Quit();
  }
  if(-2 == downstream) {	    /* if gnutella upstream */
    GnutellaEstimate(sourceIpAddress, sourcePort, targetIpAddress,
		     targetPort, &session);
    Quit();
  }
  if(-3 == downstream) {
    Latencies(sourceIpAddress, sourcePort, targetIpAddress, targetPort,
	      &session);
    Quit();
  }
  
  /* Setup connection to target */
  if ( EstablishSession(sourceIpAddress, sourcePort,
			targetIpAddress, targetPort, &session) == 0) {
    printf("ERROR: Couldn't establish session to %s:%d\n",
	   targetHostName,targetPort);
    Quit();
  }

  SendHTTPRequest(&session);

  Cleanup(1);
  close(session.socket); 
  printf("===================================================\n");
  return (0);
}
